/*
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#ifndef NOTMUCH_UTIL_PATH_UTIL_H_
#define NOTMUCH_UTIL_PATH_UTIL_H_

#ifdef __cplusplus
extern "C" {
#endif

char *
notmuch_canonicalize_file_name (const char *path);

#ifdef __cplusplus
}
#endif

#endif /* NOTMUCH_UTIL_PATH_UTIL_H_ */
