;; emacs local configuration settings for notmuch source
;; surmised by dkg on 2010-11-23 13:43:18-0500
;; amended by amdragon on 2011-06-06

((c-mode
  (indent-tabs-mode . t)
  (tab-width . 8)
  (c-basic-offset . 4)
  (c-file-style . "linux"))
 (c++-mode
  (indent-tabs-mode . t)
  (tab-width . 8)
  (c-basic-offset . 4)
  (c-file-style . "linux"))
 (emacs-lisp-mode
  (indent-tabs-mode . t)
  (tab-width . 8))
 (sh-mode
  (indent-tabs-mode . t)
  (tab-width . 8)
  (sh-basic-offset . 4)
  (sh-indentation . 4))
 (nil
  (fill-column . 70))
 )
